/* Copyright (c) 2007 Intel Corporation 
 * All rights reserved. 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 	Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 	Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     Neither the name of the Intel Corporation nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE INTEL OR ITS
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define __HEAPSAFE__
#define __HS_LIB

#include "../../include/Darwin/setjmp.h"
#include <heapsafe.h>
#include <stdlib.h>

#define JB_DVARS 19
#define JB_IVARS 20
#define JB_SCOUNT 21

static void deref_locals(void **istart, void **iend)
{
  /* Remove references due to local-variables-from-skipped-stack-frames
     which aren't handled by direct_vars (structs, arrays, variables
     whose address is taken) */
  for (; istart < iend; istart += 3)
    {
      void *obj = istart[0];
      size_t osize = (size_t)istart[1];
      hs_type_t oadj = (hs_type_t)istart[2];

      if (oadj)
	oadj(obj, -1, osize);
    }
}

static void hs_restore(sigjmp_buf env)
{
  deref_locals((void **)env[JB_IVARS], __hs_indirect_vars);

  /* Restore RC's local var info */
  __hs_indirect_vars = (void **)env[JB_IVARS];
  __hs_direct_vars = (void **)env[JB_DVARS];
  if (__hs_scount)
    {
      __hs_scount = env[JB_SCOUNT] + 1;
      hs_delayed_free_end();
    }
  else if (env[JB_SCOUNT])
    abort();
}

static void hs_save(sigjmp_buf env)
{
  env[JB_DVARS] = (int)__hs_direct_vars;
  env[JB_IVARS] = (int)__hs_indirect_vars;
  env[JB_SCOUNT] = __hs_scount;
}

int __hs_sigsetjmp(sigjmp_buf env, int val);
void __hs_siglongjmp(sigjmp_buf env, int val);

int sigsetjmp(sigjmp_buf env, int val)
{
  hs_save(env);
  return __hs_sigsetjmp(env, val);
}

void siglongjmp(sigjmp_buf env, int val)
{
  hs_restore(env);
  __hs_siglongjmp(env, val);
}


